/***********************************************//** @file HepMCVisualElements.h
*
* Header file for the graphics elements
*
* Interactive viewer for the HepMC data record: define a number of classes for
* vertices and particles that know how to draw and expand themselves.
*
* @author <A HREF="mailto:sboeser@hep.ucl.ac.uk">Sebastian B&ouml;ser</A>
* @date 16.05.07
*
*******************************************************************************/

/// Use automake header if not running in CMT
#ifndef PACKAGE_VERSION
  //-- automake header
  #include "HepMCVisual/config.h"
#endif

//-- STL libraries
#include <vector>
#include <list>
#include <algorithm>  // Include algorithms, such as sort

//-- Reflex headers
#include <typeinfo>
//#include "Reflex/Reflex.h"

//-- ROOT headers
#include "TMath.h"
#include "TLine.h"
#include "TString.h"
#include "TGFrame.h"
#include "TGLabel.h"
#include "TEllipse.h"
#include "TCurlyLine.h"
#include "TMarker.h"
#include "TPad.h"
#include "TLatex.h"
#include "TCanvas.h"
#include "TVirtualX.h"
#include "TAttText.h"
#include "Riostream.h"

//-- HepMC headers
#include "HepMC/GenParticle.h"
#include "HepMC/GenVertex.h"
#include "HepMC/GenEvent.h"

//-- Particle Data table only supported in HepMC version > 2.0
#ifndef WITH_CLHEP_VER 
  #include "HepMC/ParticleDataTable.h"
  #include "HepMC/ParticleData.h"
#endif

///Visual classes are added to same namespace as GenEvent, GenVertex and
///GenParticle
namespace HepMC {


///For compatibility with older HepMC versions, HepFourVector is defined as
/// CLHEP::HepLorentzVector for HepMC < 2.0 and
/// HepMC::FourVector for HepMC >= 2.0
#ifdef WITH_CLHEP_VER
  class HepFourVector: public CLHEP::HepLorentzVector {
    public:
      //constructor
      HepFourVector( double x, double y, double z, double t):
        HepLorentzVector(x,y,z,t) {};
    };
#else
  class HepFourVector: public HepMC::FourVector { 
    public:
      //constructor
      HepFourVector( double x, double y, double z, double t):
        FourVector(x,y,z,t) {};
    };
#endif

///Define particle types
enum EType_t {
  kOther=0, //!< any other particle
  kQuark,   //!< quarks
  kLepton,  //!< leptons
  kGauge,   //!< gauge bosons
  kMeson,   //!< meson
  kBaryon   //!< baryons
};

///Visual representation of GenParticle.
///Particles are represented by straight, wavy or curly lines. They have
///"nibbles" at both ends in order to expand/contract the tree in both
///directions. Clicking on the particle will show an info box with some
///parameters.
class VisualParticle:  public HepMC::GenParticle,
                       public TObject, public TAttLine {

  //NOTE*: This class does not publicly inherit from TLine. The reason is that
  //the user should not be able to handle lines, but only their respective
  //vertices. 
  //Curly and wavy lines for photons are only implemented via the paint method,
  //so that they can be "clicked" as lines as well.

  public:
    ///Fully qualified constructor
    VisualParticle(
      const HepMC::HepFourVector& momentum,
      int pdg_id,
      int status = 0,
      const HepMC::Flow& itsflow = Flow(0),
      const HepMC::Polarization& polar = Polarization(0.0, 0.0)
    );

    ///Constructor with a GenParticle
    VisualParticle(const HepMC::GenParticle& inparticle);
    
    ///default destructor
    ~VisualParticle();

    /** @name Get and Set functions */
    //@{
    ///Get type of particle (gauge boson/lepton/hadron/...)
    EType_t GetParticleType();
    ///Check if particles is shown (expanded)
    Bool_t GetShow(){ return fShow; };
    ///Set particle to be shown
    void SetShow(Bool_t show=kTRUE){ fShow = show; };
    ///Check if particle shall be reconnected if redrawn
    Bool_t GetReconnect(){ return fReconnect; };
    ///Set particle to be reconnected on redraw
    void SetReconnect(Bool_t connect=kTRUE){ fReconnect = connect; };
    ///Check if particle is currently drawn on the canvas
    Bool_t GetDrawn();
    ///Set particle line attributes to default
    void SetDefaultStyle();
    ///Get default length for particles with only one end vertex
    Double_t GetDefaultLength() { return fDefaultLength; };
    ///Get number of siblings (particles inbetween same vertices)
    std::vector<VisualParticle*> GetSiblings();
    ///Check if particles label is always shown
    Bool_t GetStaticLabel(){ return fStaticLabel; };
    ///Set particle to be shown
    void SetStaticLabel(Bool_t isStatic=kTRUE){ fStaticLabel = isStatic; };
    //@}

    /** @name Graphics get and Set functions */
    //@{
    ///Get X-coordinate of starting point (NDC)
    Double_t GetStartX();               //Start position from vertex
    ///Get Y-coordinate of starting point (NDC)
    Double_t GetStartY();               //Start position from vertex
    ///Get X-coordinate of end point (NDC)
    Double_t GetEndX();                 //End position from vertex
    ///Get Y-coordinate of end point (NDC)
    Double_t GetEndY();                 //End position from vertex
    ///Get angle of particle line - relative to vertically upwards
    Double_t GetDir();                  //theta angle of particle
    //@}
    
    /** @name Graphics methods */
    //@{
    ///Get distance from mouse pointer - closest object is picked.
    Int_t DistancetoPrimitive(Int_t px, Int_t py);
    ///Handle mouse events
    void ExecuteEvent(Int_t event, Int_t px, Int_t py);
    ///Add particle to the canvas
    void Draw(Option_t* option = "");
    ///Initialize other end point of particles with only one vertex
    Bool_t InitializeLine();
    ///Remove particle from the canvas
    void Undraw();
    ///Check all parents and downward for (re)draw
    void UpdateParents();
    ///Check all children and upwards for (re)draw
    void UpdateChildren();
    ///Called each time another object is removed from the canvas. Check for
    ///lost vertices.
    void RecursiveRemove(TObject* obj);
    ///Do the actual paint job - calls PaintLine(),
    ///PaintStartNibble(), PaintEndNibble()
    void Paint(Option_t* option = "");
    ///Paint the line
    void PaintLine(Option_t* option = "");
    ///Paint the nibble at the production vertex
    void PaintStartNibble(Option_t* option = ""); 
    ///Paint the nibble at the end vertex
    void PaintEndNibble(Option_t* option = "");
    ///Paint the name of the particle if mouse if moved over it
    void PaintLabel(Int_t px, Int_t py);
    ///Assemble the particle info box
    void MakeInfoFrame();
    ///Get the particle info box
    TGCompositeFrame* GetInfoFrame();
    ///Show this particles info box
    void ShowInfoFrame(Int_t px, Int_t py);
    //@}


  private:

    ///Default initialization for all constructors
    void Initialize();       
    ///Returns either of the vertices
    GenVertex* AnyVertex();

    /** @name Graphics flags */
    //@{
    Bool_t fShow;            //!<Wether the particle is displayed or not
    Bool_t fReconnect;       //!<Reconnect to graph when draw is called
    Bool_t fMovingLine;      //!<Wether in LineMoving mode (no vertex)
    Bool_t fStaticLabel;     //!<Wether the particles labels is always shown
    //@}

    /** @name Graphics elements */
    //@{
    TGCompositeFrame* fInfoFrame;      //!<Popup Frame for particle infos
    TAttText fAttLabel;                //!<Particle Name/ID label attributes
    TLine fLine;                       //!<All particles are lines
    TEllipse fStartNibble;             //!<Nibble at production vertex
    TEllipse fEndNibble;               //!<Nibble at end vertex
    Double_t fNibbleXRad, fNibbleYRad; //!<Nibble dimensions
    Double_t fDefaultLength;           //!<Default particle length
    //@}

};

///Visual representation of GenVertex.
///VisualVertex directly derives from GenVertex and TMarker. It can hold both,
///GenParticles and VisualParticles. If drawn on a canvas, the GenParticles will
///be automatically converted. The TMarker representing the vertex can be moved
///around on the canvas. On double-click, all the parent and daughter particles
///will be shown.
class VisualVertex: public TMarker , public HepMC::GenVertex {

  public:
    ///Fully qualified constructor
    VisualVertex(
      Double_t x,
      Double_t y,
      const HepMC::HepFourVector& position = HepFourVector(0.0, 0.0, 0.0, 0.0),
      int id = 0,
      const HepMC::WeightContainer& weights = std::vector<double>()
    );
    
    ///Constructor with a GenVertex
    VisualVertex(Double_t x, Double_t y, const HepMC::GenVertex& invertex);

    ///Default constructor
    ~VisualVertex();

    /** @name Get and set methods */
    //@{
    ///Check if vertex is drawn on the canvas
    Bool_t GetDrawn();
    ///Check if vertex is protected from undrawing. This avoids removal of the
    ///of the full graph in case of loops.
    Bool_t GetProtect(){return fProtect;};
    ///Set protection state of vertex.
    void SetProtect(Bool_t protect=kTRUE){fProtect = protect;};
    ///Get number of shown children
    Int_t GetShownChildren();
    ///Get number of shown parents
    Int_t GetShownParents();
    ///Get a direction for a new parent particle.
    Double_t GetNewParentDir();
    ///Get a direction for a new child particle.
    Double_t GetNewChildDir();
    //@}
    
    /** @name Graphics methods */
    //@{
    ///Get distance from mouse pointer - closest object is picked.
    Int_t DistancetoPrimitive(Int_t px, Int_t py);
    ///Handle mouse events
    void ExecuteEvent(Int_t event, Int_t px, Int_t py);
    ///Add particle to the canvas
    void Draw(Option_t* option = "");
    ///Do the actual paint job - overriden to check for changed particles
    void Paint(Option_t* option = "");
    ///Paint the vertex label if mouse is moved over
    void PaintLabel(Int_t px, Int_t py);
    ///Remove vertex from canvas
    void Undraw();
    ///Collapse all parents from this vertex upwards
    void UndrawParents();
    ///Collapse all children from this vertex downwards
    void UndrawChildren();
    ///Unshow or show all the particles at this vertex
    void ToggleAll();
    //@}

  private:
    ///Default initialization for all constructors
    void Initialize();
    
    Bool_t fProtect;                    //!<Avoid undraws from loops
    TAttText fAttLabel;                 //!<Vertex label text attributes

};

///Container for VisualVertex and GenVertex members.
///VisualEvent directly derives from GenEvent, but can use
///visual represenatations of the GenVertex and GenParticle
///classes. It also inherits from TObject for support of the Draw functionality.
///\em Note: A VisualEvent can hold both, GenVertices and VisualVertices.
///Conversion of one into the other will be taken care of by the Draw functions.
class VisualEvent: public HepMC::GenEvent, public TObject {

  public:
    ///Fully qualified constructor
    VisualEvent(
      int signal_process_id = 0,
      int event_number = 0,
      HepMC::GenVertex* signal_vertex = 0,
      const HepMC::WeightContainer& weights = std::vector<double>(),
      const vector<long>& randomstates = std::vector<long int>()
//      HepMC::HeavyIon& ion, HepMC::PdfInfo& pdf
    );

    ///Constructor with a GenEvent.
    ///This will create a complete copy of the event, but will not replace the
    ///GenVertices and GenParticles by VisualVertex and VisualParticle.
    VisualEvent(const HepMC::GenEvent& inevent);

//Particle data tables not supported by CLHEP version
#ifndef WITH_CLHEP_VER
    /** @name Get and Set methods */
    //@{
    ///Set reference to a particle data table
    void SetPDT( const ParticleDataTable* pdt ) { fpdt = pdt ; };
    ///Get reference to the particle data table
    const ParticleDataTable* GetPDT() { return fpdt ; };
    ///Shortcut to get particle data from table
    ParticleData* GetParticleData(Int_t pid = -1);
    ///Shortcut to get particle name from table
    std::string GetParticleName(Int_t pid = -1);
    //@}
#endif
    
    /** @name Graphics methods */
    //@{
    ///Draw the signal vertex. If not set, the first vertex from the event is
    ///drawn instead.
    void Draw(Option_t* option = "");
    //@}

    ///Print event summary as text. This prints the event using the
    ///GenEvent::print() function. As VisualEvent globally inherits from
    ///GenEvent, one might as well use GenEvent::print() in C++. However, this
    ///will not work in python, as the print statement is reserved in python.
    void print_event(ostream& ostr = std::cout) { print(ostr); };

  private:

//Particle data tables not supported by CLHEP version
#ifndef WITH_CLHEP_VER
    const ParticleDataTable* fpdt;               //!<A reference to a PDT Table
#endif
};

///@cond NEVER
// Exclude this bit from doxygen documentation
//using namespace ROOT::Reflex ;

//static definition of class types speeds up frequent lookup
static VisualParticle VisParticleType(HepFourVector(0,0,0,0),0,0,0,0);
static VisualVertex VisVertexType(-1,-1);
static VisualEvent VisEventType(0,0);
//static Type VisParticleType = Type::ByName("HepMC::VisualParticle");
//static Type VisVertexType = Type::ByName("HepMC::VisualVertex");
//static Type VisEventType = Type::ByName("HepMC::VisualEvent");

///@endcond

///Check if object is one of the visual classes. This class can be used with
///the HepMC::copy_if function.
class Is_Visual {
  public:
    bool operator() ( GenVertex* invertex );
    bool operator() ( GenParticle* inparticle );
    bool operator() ( GenEvent* inevent );
};


/// This static instance of the Is_Visual class can be used to check if any
/// object is of VisualEvent, VisualVertex or VisualParticle type
static Is_Visual IsVisual;

/// Check if object is \b NOT one of the visual classes. This class can be used
/// with the HepMC::copy_if function.
class Is_Not_Visual {
  public:
    bool operator()(GenVertex* invertex ){ return ! IsVisual(invertex); };
    bool operator()(GenParticle* inparticle){ return ! IsVisual(inparticle); };
    bool operator()(GenEvent* inevent){ return ! IsVisual(inevent); };
};

/** @name Global conversion functions */
//@{

/// Simple global conversion functions
VisualParticle* Visualize(HepMC::GenParticle* inparticle);
VisualVertex* Visualize(HepMC::GenVertex* invertex);
VisualEvent* Visualize(HepMC::GenEvent* inevent);
//@}

} // end of namespace HepMC
